/*
 * test_validator.c
 *
 * Copyright 2023 Jonathan Schöbel <jonathan@xn--schbel-yxa.info>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <check.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>


/* lower SIZE_MAX as we try to reach it */
#include <limits.h>
#undef SIZE_MAX
#define SIZE_MAX 10 * sizeof (struct tag_info)

#include "macro.h"
#include "status.h"

/* C file is needed, because we want to override SIZE_MAX */
#include "validator.c"


START_TEST(test_validator_no_status)
{
	struct SH_Validator * validator;

	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;

	_status_preinit (status);
	validator = SH_Validator_new (&status);
	ck_assert_ptr_ne (NULL, validator);
	ck_assert_int_eq (status.status, SUCCESS);

	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_copy_no_status)
{
	struct SH_Validator * validator;
	struct SH_Validator * copy;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test */
	copy = SH_Validator_copy (validator, NULL);
	ck_assert_ptr_ne (NULL, copy);

	/* cleanup */
	SH_Validator_free (copy);
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_copy_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;
	struct SH_Validator * copy;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test */
	_status_preinit (status);
	copy = SH_Validator_copy (validator, &status);
	ck_assert_ptr_ne (NULL, copy);
	ck_assert_int_eq (SUCCESS, status.status);

	/* cleanup */
	SH_Validator_free (copy);
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_spec_no_status)
{
	struct SH_Validator * validator;

	/* test */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_spec_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;

	/* test */
	_status_preinit (status);
	validator = SH_Validator_new_html5 (&status);
	ck_assert_ptr_ne (NULL, validator);
	ck_assert_int_eq (SUCCESS, status.status);

	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

Suite * validator_suite (void)
{
	Suite *s;
	TCase *tc_core;

	s = suite_create ("Testsuite SeFHT Validator");

	/* Core test case */
	tc_core = tcase_create ("Core");

	tcase_add_test (tc_core, test_validator_no_status);
	tcase_add_test (tc_core, test_validator_with_status);
	tcase_add_test (tc_core, test_validator_copy_no_status);
	tcase_add_test (tc_core, test_validator_copy_with_status);
	tcase_add_test (tc_core, test_validator_spec_no_status);
	tcase_add_test (tc_core, test_validator_spec_with_status);
	suite_add_tcase (s, tc_core);

	return s;
}

int main (void)
{
	int number_failed;
	Suite *s;
	SRunner *sr;

	s = validator_suite ();
	sr = srunner_create (s);

	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);

	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

