/*
 * test_validator_attr.c
 *
 * Copyright 2023 Jonathan Schöbel <jonathan@xn--schbel-yxa.info>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <check.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>


/* lower SIZE_MAX as we try to reach it */
#include <limits.h>
#undef SIZE_MAX
#define SIZE_MAX 10 * sizeof (struct attr_info)

#include "macro.h"
#include "status.h"


/* C file is needed, because we want to override SIZE_MAX */
#include "validator.c"


START_TEST(test_validator_no_status)
{
	struct SH_Validator * validator;

	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	ck_assert_int_eq (0, validator->attr_n);

	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;

	_status_preinit (status);
	validator = SH_Validator_new (&status);
	ck_assert_ptr_ne (NULL, validator);
	ck_assert_int_eq (status.status, SUCCESS);

	ck_assert_int_eq (0, validator->attr_n);

	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_copy_no_status)
{
	struct SH_Validator * validator;
	struct SH_Validator * copy;
	bool result;

	/* setup */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	result = SH_Validator_register_attr (validator, "body", "id",
	                                     NULL);
	ck_assert_int_ne (FALSE, result);

	result = SH_Validator_register_attr (validator, "html", "lang",
	                                     NULL);
	ck_assert_int_ne (FALSE, result);

	/* test */
	copy = SH_Validator_copy (validator, NULL);
	ck_assert_ptr_ne (NULL, copy);

	ck_assert_ptr_ne (NULL, copy->attrs);
	ck_assert_ptr_ne (validator->attrs, copy->attrs);
	ck_assert_int_eq (validator->attr_n, copy->attr_n);

	#define TEST_INT(I) ck_assert_int_eq (validator->I, copy->I);
	#define TEST_PTR(P) ck_assert_ptr_eq (validator->P, copy->P);
	#define TEST_STR(S) ck_assert_ptr_ne (validator->S, copy->S);  \
	                    ck_assert_str_eq (validator->S, copy->S);

	for (size_t index = 0; index < copy->attr_n; index++)
	{
		TEST_STR(attrs[index].name)

		TEST_INT(attrs[index].tag_n)
		for (size_t j = 0; j < copy->attrs[index].tag_n; j++)
		{
			TEST_PTR(attrs[index].tags[j].name)
		}
	}
	#undef TEST_INT
	#undef TEST_PTR
	#undef TEST_STR

	SH_Validator_free (copy);
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_copy_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;
	struct SH_Validator * copy;
	bool result;

	/* setup */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	result = SH_Validator_register_attr (validator, "body", "id",
	                                     NULL);
	ck_assert_int_ne (FALSE, result);

	result = SH_Validator_register_attr (validator, "html", "lang",
	                                     NULL);
	ck_assert_int_ne (FALSE, result);

	/* test */
	_status_preinit (status);
	copy = SH_Validator_copy (validator, &status);
	ck_assert_ptr_ne (NULL, copy);
	ck_assert_int_eq (status.status, SUCCESS);

	ck_assert_ptr_ne (NULL, copy->attrs);
	ck_assert_ptr_ne (validator->attrs, copy->attrs);
	ck_assert_int_eq (validator->attr_n, copy->attr_n);

	#define TEST_INT(I) ck_assert_int_eq (validator->I, copy->I);
	#define TEST_PTR(P) ck_assert_ptr_eq (validator->P, copy->P);
	#define TEST_STR(S) ck_assert_ptr_ne (validator->S, copy->S);  \
	                    ck_assert_str_eq (validator->S, copy->S);

	for (size_t index = 0; index < copy->attr_n; index++)
	{
		TEST_STR(attrs[index].name)

		TEST_INT(attrs[index].tag_n)
		for (size_t j = 0; j < copy->attrs[index].tag_n; j++)
		{
			TEST_PTR(attrs[index].tags[j].name)
		}
	}
	#undef TEST_INT
	#undef TEST_PTR
	#undef TEST_STR

	SH_Validator_free (copy);
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_register_no_status)
{
	struct SH_Validator * validator;
	const char * tag1 = "html";
	const char * tag2 = "foobarbaz"; // mustn't be registered yet
	const char * tag3 = "body";
	const char * tag4 = "nav";
	const char * tag5 = "aside";
	const char * attr1 = "id";
	const char * attr2 = "class";
	const char * attr3 = "name";
	const char * attr4 = "src";
	const char * attr5 = "content";
	char * attrN;
	bool result;

	/* setup */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test - register */
	result = SH_Validator_register_attr (validator, tag1, attr1,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	ck_assert_int_eq (1, validator->attr_n);
	ck_assert_ptr_ne (attr1, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[0].name);

	ck_assert_int_eq (1, validator->attrs[0].tag_n);
	ck_assert_ptr_ne (tag1, validator->attrs[0].tags[0].name);
	ck_assert_str_eq (tag1, validator->attrs[0].tags[0].name);

	/* test - duplicate registration */
	result = SH_Validator_register_attr (validator, tag1, attr1,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	ck_assert_int_eq (1, validator->attr_n);
	ck_assert_ptr_ne (attr1, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[0].name);

	ck_assert_int_eq (1, validator->attrs[0].tag_n);
	ck_assert_ptr_ne (tag1, validator->attrs[0].tags[0].name);
	ck_assert_str_eq (tag1, validator->attrs[0].tags[0].name);

	/* test - automatic tag register */
	result = SH_Validator_register_attr (validator, tag2, attr1,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	ck_assert_int_eq (1, validator->attr_n);
	ck_assert_ptr_ne (attr1, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[0].name);

	#define TEST_STR(S1, S2) ck_assert_ptr_ne (S1, S2);            \
	                         ck_assert_str_eq (S1, S2);

	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	/* The storage order depends on the relative position of memory,
	 * allocated by different malloc calls. This can change and
	 * thus must be determined. */
	if (validator->attrs[0].tags[0].name[0]
	  > validator->attrs[0].tags[1].name[0])
	{
		TEST_STR(tag1, validator->attrs[0].tags[0].name)
		TEST_STR(tag2, validator->attrs[0].tags[1].name)
	}
	else
	{
		TEST_STR(tag2, validator->attrs[0].tags[0].name)
		TEST_STR(tag1, validator->attrs[0].tags[1].name)
	}

	#undef TEST_STR

	/* test - order (attr) */
	result = SH_Validator_register_attr (validator, tag1, attr3,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_register_attr (validator, tag1, attr4,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_register_attr (validator, tag1, attr5,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	ck_assert_int_eq (4, validator->attr_n);
	ck_assert_str_eq (attr5, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[1].name);
	ck_assert_str_eq (attr3, validator->attrs[2].name);
	ck_assert_str_eq (attr4, validator->attrs[3].name);

	/* test - order (tag) */
	result = SH_Validator_register_attr (validator, tag2, attr4,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_register_attr (validator, tag3, attr4,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_register_attr (validator, tag4, attr4,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_register_attr (validator, tag5, attr4,
	                                     NULL);
	ck_assert_int_eq (TRUE, result);

	#define attr validator->attrs[3]
	ck_assert_int_eq (5, attr.tag_n);
	for (size_t i = 0; i < attr.tag_n-1; i++)
	{
		ck_assert_int_lt ((size_t)attr.tags[i].name,
		                  (size_t)attr.tags[i+1].name);
	}
	#undef attr

	/* test - overflow detection */
	/* make method fail by filling with garbage until
	 * upper boundary is reached */

	/* ensure enough space  inside string*/
	/* log10 +1 = number length */
	/* +4 "attr" */
	/* +1 '\0' */
	/* = +5 */
	attrN = calloc (((int) floor (log10 ((double) SIZE_MAX))) + 6,
		       sizeof (char));

	/* fill with garbage */
	do
	{
		sprintf (attrN, "attr%zu", validator->attr_n);
	}
	while (SH_Validator_register_attr (validator, tag1, attrN, NULL));

	free (attrN);

	/* test overflow */
	result = SH_Validator_register_attr (validator, tag1, attr2,
	                                     NULL);
	ck_assert_int_eq (FALSE, result);

	ck_assert_int_eq (10, validator->attr_n);

	/* cleanup */
	/* also free garbage created for overflow test */
	validator->attr_n = 10;
	SH_Validator_free (validator);
}

START_TEST(test_validator_register_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;
	const char * tag1 = "html";
	const char * tag2 = "foobarbaz"; // mustn't be registered yet
	const char * tag3 = "body";
	const char * tag4 = "nav";
	const char * tag5 = "aside";
	const char * attr1 = "id";
	const char * attr2 = "class";
	const char * attr3 = "name";
	const char * attr4 = "src";
	const char * attr5 = "content";
	char * attrN;
	bool result;

	/* setup */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test - register */
	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag1, attr1,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (1, validator->attr_n);
	ck_assert_ptr_ne (attr1, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[0].name);

	ck_assert_int_eq (1, validator->attrs[0].tag_n);
	ck_assert_ptr_ne (tag1, validator->attrs[0].tags[0].name);
	ck_assert_str_eq (tag1, validator->attrs[0].tags[0].name);

	/* test - duplicate registration */
	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag1, attr1,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (1, validator->attr_n);
	ck_assert_ptr_ne (attr1, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[0].name);

	ck_assert_int_eq (1, validator->attrs[0].tag_n);
	ck_assert_ptr_ne (tag1, validator->attrs[0].tags[0].name);
	ck_assert_str_eq (tag1, validator->attrs[0].tags[0].name);

	/* test - automatic tag register */
	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag2, attr1,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (1, validator->attr_n);
	ck_assert_ptr_ne (attr1, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[0].name);

	#define TEST_STR(S1, S2) ck_assert_ptr_ne (S1, S2);            \
	                         ck_assert_str_eq (S1, S2);

	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	/* The storage order depends on the relative position of memory,
	 * allocated by different malloc calls. This can change and
	 * thus must be determined. */
	if (validator->attrs[0].tags[0].name[0]
	  > validator->attrs[0].tags[1].name[0])
	{
		TEST_STR(tag1, validator->attrs[0].tags[0].name)
		TEST_STR(tag2, validator->attrs[0].tags[1].name)
	}
	else
	{
		TEST_STR(tag2, validator->attrs[0].tags[0].name)
		TEST_STR(tag1, validator->attrs[0].tags[1].name)
	}

	#undef TEST_STR

	/* test - order (attr) */
	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag1, attr3,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag1, attr4,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag1, attr5,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (4, validator->attr_n);
	ck_assert_str_eq (attr5, validator->attrs[0].name);
	ck_assert_str_eq (attr1, validator->attrs[1].name);
	ck_assert_str_eq (attr3, validator->attrs[2].name);
	ck_assert_str_eq (attr4, validator->attrs[3].name);

	/* test - order (tag) */
	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag2, attr4,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag3, attr4,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag4, attr4,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag5, attr4,
	                                     &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (5, validator->attrs[3].tag_n);

	#define attr validator->attrs[3]
	ck_assert_int_eq (5, attr.tag_n);
	for (size_t i = 0; i < attr.tag_n-1; i++)
	{
		ck_assert_int_lt ((size_t)attr.tags[i].name,
		                  (size_t)attr.tags[i+1].name);
	}
	#undef attr

	/* test - overflow detection */
	/* make method fail by filling with garbage until
	 * upper boundary is reached */

	/* ensure enough space  inside string*/
	/* log10 +1 = number length */
	/* +4 "attr" */
	/* +1 '\0' */
	/* = +5 */
	attrN = calloc (((int) floor (log10 ((double) SIZE_MAX))) + 6,
		       sizeof (char));

	/* fill with garbage */
	do
	{
		sprintf (attrN, "attr%zu", validator->attr_n);
	}
	while (SH_Validator_register_attr (validator, tag1, attrN, NULL));

	free (attrN);

	/* test overflow */
	_status_preinit (status);
	result = SH_Validator_register_attr (validator, tag1, attr2,
	                                     &status);
	ck_assert_int_eq (FALSE, result);
	ck_assert_int_eq (E_DOMAIN, status.status);

	ck_assert_int_eq (10, validator->attr_n);

	/* cleanup */
	/* also free garbage created for overflow test */
	validator->attr_n = 10;
	SH_Validator_free (validator);
}

START_TEST(test_validator_deregister_no_status)
{
	struct SH_Validator * validator;
	bool result;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	#define REGISTER SH_Validator_register_attr
	result = REGISTER (validator, "html", "attr", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "html", "id", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "html", "name", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "html", "class", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "attr", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "id", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "name", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "class", NULL);
	ck_assert_int_ne (FALSE, result);
	#undef REGISTER

	/* test - consistency */
	ck_assert_int_eq (4, validator->attr_n);
	ck_assert_str_eq ("attr", validator->attrs[0].name);
	ck_assert_str_eq ("class", validator->attrs[1].name);
	ck_assert_str_eq ("id", validator->attrs[2].name);
	ck_assert_str_eq ("name", validator->attrs[3].name);
	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	ck_assert_int_eq (2, validator->attrs[1].tag_n);
	ck_assert_int_eq (2, validator->attrs[2].tag_n);
	ck_assert_int_eq (2, validator->attrs[3].tag_n);

	#define STR_EQ ck_assert_str_eq
	/* The storage order depends on the relative position of memory,
	 * allocated by different malloc calls. This can change and
	 * thus must be determined. */
	if (validator->attrs[0].tags[0].name[0]
	  > validator->attrs[0].tags[1].name[0])
	{
		STR_EQ ("html", validator->attrs[0].tags[0].name);
		STR_EQ ("html", validator->attrs[1].tags[0].name);
		STR_EQ ("html", validator->attrs[2].tags[0].name);
		STR_EQ ("html", validator->attrs[3].tags[0].name);
		STR_EQ ("body", validator->attrs[0].tags[1].name);
		STR_EQ ("body", validator->attrs[1].tags[1].name);
		STR_EQ ("body", validator->attrs[2].tags[1].name);
		STR_EQ ("body", validator->attrs[3].tags[1].name);
	}
	else
	{
		STR_EQ ("html", validator->attrs[0].tags[1].name);
		STR_EQ ("html", validator->attrs[1].tags[1].name);
		STR_EQ ("html", validator->attrs[2].tags[1].name);
		STR_EQ ("html", validator->attrs[3].tags[1].name);
		STR_EQ ("body", validator->attrs[0].tags[0].name);
		STR_EQ ("body", validator->attrs[1].tags[0].name);
		STR_EQ ("body", validator->attrs[2].tags[0].name);
		STR_EQ ("body", validator->attrs[3].tags[0].name);
	}
	#undef STR_EQ

	/* test - existent attr */
	result = SH_Validator_deregister_attr (validator,
	                                       "html", "attr", NULL);
	ck_assert_int_eq (TRUE, result);

	/* test - non existent tag */
	result = SH_Validator_deregister_attr (validator,
	                                       "html", "attr", NULL);
	ck_assert_int_eq (FALSE, result);

	/* test - existent attr, auto remove */
	result = SH_Validator_deregister_attr (validator,
	                                       "body", "attr", NULL);
	ck_assert_int_eq (TRUE, result);

	/* test - consistency */
	ck_assert_int_eq (3, validator->attr_n);
	ck_assert_str_eq ("class", validator->attrs[0].name);
	ck_assert_str_eq ("id", validator->attrs[1].name);
	ck_assert_str_eq ("name", validator->attrs[2].name);
	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	ck_assert_int_eq (2, validator->attrs[1].tag_n);
	ck_assert_int_eq (2, validator->attrs[2].tag_n);


	/* test - non existent attr */
	result = SH_Validator_deregister_attr (validator, "html", "attr",
	                                       NULL);
	ck_assert_int_eq (FALSE, result);

	/* test - existent attr, total remove */
	result = SH_Validator_deregister_attr (validator, NULL, "name",
	                                       NULL);
	ck_assert_int_eq (TRUE, result);

	/* test - non existent attr, total remove */
	result = SH_Validator_deregister_attr (validator, NULL, "attr",
	                                       NULL);
	ck_assert_int_eq (FALSE, result);

	/* test - consistency */
	ck_assert_int_eq (2, validator->attr_n);
	ck_assert_str_eq ("class", validator->attrs[0].name);
	ck_assert_str_eq ("id", validator->attrs[1].name);
	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	ck_assert_int_eq (2, validator->attrs[1].tag_n);


	#define STR_EQ ck_assert_str_eq
	/* The storage order depends on the relative position of memory,
	 * allocated by different malloc calls. This can change and
	 * thus must be determined. */
	if (validator->attrs[0].tags[0].name[0]
	  > validator->attrs[0].tags[1].name[0])
	{
		STR_EQ ("html", validator->attrs[0].tags[0].name);
		STR_EQ ("html", validator->attrs[1].tags[0].name);
		STR_EQ ("body", validator->attrs[0].tags[1].name);
		STR_EQ ("body", validator->attrs[1].tags[1].name);
	}
	else
	{
		STR_EQ ("html", validator->attrs[0].tags[1].name);
		STR_EQ ("html", validator->attrs[1].tags[1].name);
		STR_EQ ("body", validator->attrs[0].tags[0].name);
		STR_EQ ("body", validator->attrs[1].tags[0].name);
	}
	#undef STR_EQ


	/* test - existent tag, total remove */
	result = SH_Validator_deregister_attr (validator, "html", NULL,
	                                       NULL);
	ck_assert_int_eq (TRUE, result);

	/* test - non existent tag, total remove */
	result = SH_Validator_deregister_attr (validator, "nav", NULL,
	                                       NULL);
	ck_assert_int_eq (FALSE, result);

	/* test - consistency */
	ck_assert_int_eq (2, validator->attr_n);
	ck_assert_str_eq ("class", validator->attrs[0].name);
	ck_assert_str_eq ("id", validator->attrs[1].name);
	ck_assert_int_eq (1, validator->attrs[0].tag_n);
	ck_assert_int_eq (1, validator->attrs[1].tag_n);
	ck_assert_str_eq ("body", validator->attrs[0].tags[0].name);
	ck_assert_str_eq ("body", validator->attrs[1].tags[0].name);

	/* test - total clear */
	result = SH_Validator_deregister_attr (validator, NULL, NULL,
	                                       NULL);
	ck_assert_int_eq (TRUE, result);

	/* test - consistency */
	ck_assert_int_eq (0, validator->attr_n);


	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_deregister_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;
	bool result;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	#define REGISTER SH_Validator_register_attr
	result = REGISTER (validator, "html", "attr", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "html", "id", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "html", "name", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "html", "class", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "attr", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "id", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "name", NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, "body", "class", NULL);
	ck_assert_int_ne (FALSE, result);
	#undef REGISTER

	/* test - consistency */
	ck_assert_int_eq (4, validator->attr_n);
	ck_assert_str_eq ("attr", validator->attrs[0].name);
	ck_assert_str_eq ("class", validator->attrs[1].name);
	ck_assert_str_eq ("id", validator->attrs[2].name);
	ck_assert_str_eq ("name", validator->attrs[3].name);
	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	ck_assert_int_eq (2, validator->attrs[1].tag_n);
	ck_assert_int_eq (2, validator->attrs[2].tag_n);
	ck_assert_int_eq (2, validator->attrs[3].tag_n);

	#define STR_EQ ck_assert_str_eq
	/* The storage order depends on the relative position of memory,
	 * allocated by different malloc calls. This can change and
	 * thus must be determined. */
	if (validator->attrs[0].tags[0].name[0]
	  > validator->attrs[0].tags[1].name[0])
	{
		STR_EQ ("html", validator->attrs[0].tags[0].name);
		STR_EQ ("html", validator->attrs[1].tags[0].name);
		STR_EQ ("html", validator->attrs[2].tags[0].name);
		STR_EQ ("html", validator->attrs[3].tags[0].name);
		STR_EQ ("body", validator->attrs[0].tags[1].name);
		STR_EQ ("body", validator->attrs[1].tags[1].name);
		STR_EQ ("body", validator->attrs[2].tags[1].name);
		STR_EQ ("body", validator->attrs[3].tags[1].name);
	}
	else
	{
		STR_EQ ("html", validator->attrs[0].tags[1].name);
		STR_EQ ("html", validator->attrs[1].tags[1].name);
		STR_EQ ("html", validator->attrs[2].tags[1].name);
		STR_EQ ("html", validator->attrs[3].tags[1].name);
		STR_EQ ("body", validator->attrs[0].tags[0].name);
		STR_EQ ("body", validator->attrs[1].tags[0].name);
		STR_EQ ("body", validator->attrs[2].tags[0].name);
		STR_EQ ("body", validator->attrs[3].tags[0].name);
	}
	#undef STR_EQ


	/* test - existent attr */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, "html", "attr",
	                                       &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	/* test - non existent tag */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, "html", "attr",
	                                       &status);
	ck_assert_int_eq (FALSE, result);
	ck_assert_int_eq (E_VALUE, status.status);

	/* test - existent attr, auto remove */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, "body", "attr",
	                                       &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	/* test - consistency */
	ck_assert_int_eq (3, validator->attr_n);
	ck_assert_str_eq ("class", validator->attrs[0].name);
	ck_assert_str_eq ("id", validator->attrs[1].name);
	ck_assert_str_eq ("name", validator->attrs[2].name);
	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	ck_assert_int_eq (2, validator->attrs[1].tag_n);
	ck_assert_int_eq (2, validator->attrs[2].tag_n);


	/* test - non existent attr */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, "html", "attr",
	                                       &status);
	ck_assert_int_eq (FALSE, result);
	ck_assert_int_eq (E_VALUE, status.status);

	/* test - existent attr, total remove */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, NULL, "name",
	                                       &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	/* test - non existent attr, total remove */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, NULL, "attr",
	                                       &status);
	ck_assert_int_eq (FALSE, result);
	ck_assert_int_eq (E_VALUE, status.status);

	/* test - consistency */
	ck_assert_int_eq (2, validator->attr_n);
	ck_assert_str_eq ("class", validator->attrs[0].name);
	ck_assert_str_eq ("id", validator->attrs[1].name);
	ck_assert_int_eq (2, validator->attrs[0].tag_n);
	ck_assert_int_eq (2, validator->attrs[1].tag_n);


	#define STR_EQ ck_assert_str_eq
	/* The storage order depends on the relative position of memory,
	 * allocated by different malloc calls. This can change and
	 * thus must be determined. */
	if (validator->attrs[0].tags[0].name[0]
	  > validator->attrs[0].tags[1].name[0])
	{
		STR_EQ ("html", validator->attrs[0].tags[0].name);
		STR_EQ ("html", validator->attrs[1].tags[0].name);
		STR_EQ ("body", validator->attrs[0].tags[1].name);
		STR_EQ ("body", validator->attrs[1].tags[1].name);
	}
	else
	{
		STR_EQ ("html", validator->attrs[0].tags[1].name);
		STR_EQ ("html", validator->attrs[1].tags[1].name);
		STR_EQ ("body", validator->attrs[0].tags[0].name);
		STR_EQ ("body", validator->attrs[1].tags[0].name);
	}
	#undef STR_EQ


	/* test - existent tag, total remove */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, "html", NULL,
	                                       &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	/* test - non existent tag, total remove */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, "nav", NULL,
	                                       &status);
	ck_assert_int_eq (FALSE, result);
	ck_assert_int_eq (E_VALUE, status.status);

	/* test - consistency */
	ck_assert_int_eq (2, validator->attr_n);
	ck_assert_str_eq ("class", validator->attrs[0].name);
	ck_assert_str_eq ("id", validator->attrs[1].name);
	ck_assert_int_eq (1, validator->attrs[0].tag_n);
	ck_assert_int_eq (1, validator->attrs[1].tag_n);
	ck_assert_str_eq ("body", validator->attrs[0].tags[0].name);
	ck_assert_str_eq ("body", validator->attrs[1].tags[0].name);

	/* test - total clear */
	_status_preinit (status);
	result = SH_Validator_deregister_attr (validator, NULL, NULL,
	                                       &status);
	ck_assert_int_eq (TRUE, result);
	ck_assert_int_eq (SUCCESS, status.status);

	/* test - consistency */
	ck_assert_int_eq (0, validator->attr_n);


	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_check)
{
	struct SH_Validator * validator;
	const char * tag1 = "html";
	const char * tag2 = "html";
	const char * tag3 = "head";
	const char * tag4 = "body";
	const char * attr1  = "id";
	const char * attr2  = "id";
	const char * attr3  = "class";
	const char * attr4  = "name";
	bool result;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);


	#define REGISTER SH_Validator_register_attr
	result = REGISTER (validator, tag1, attr1, NULL);
	ck_assert_int_ne (FALSE, result);
	result = REGISTER (validator, tag3, attr3, NULL);
	ck_assert_int_ne (FALSE, result);
	#undef REGISTER

	/* test */
	result = SH_Validator_check_attr (validator, tag1, attr1);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_check_attr (validator, tag2, attr2);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_check_attr (validator, tag1, attr3);
	ck_assert_int_eq (FALSE, result);

	result = SH_Validator_check_attr (validator, tag3, attr4);
	ck_assert_int_eq (FALSE, result);

	result = SH_Validator_check_attr (validator, tag4, attr2);
	ck_assert_int_eq (FALSE, result);

	result = SH_Validator_check_attr (validator, tag4, attr4);
	ck_assert_int_eq (FALSE, result);

	result = SH_Validator_check_attr (validator, NULL, attr2);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_check_attr (validator, NULL, attr4);
	ck_assert_int_eq (FALSE, result);

	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

Suite * test_suite (void)
{
	Suite *s;
	TCase *tc_core;

	s = suite_create ("Testsuite SeFHT Validator Tag");

	/* Core test case */
	tc_core = tcase_create ("Core");

	tcase_add_test (tc_core, test_validator_no_status);
	tcase_add_test (tc_core, test_validator_with_status);
	tcase_add_test (tc_core, test_validator_copy_no_status);
	tcase_add_test (tc_core, test_validator_copy_with_status);
	tcase_add_test (tc_core, test_validator_register_no_status);
	tcase_add_test (tc_core, test_validator_register_with_status);
	tcase_add_test (tc_core, test_validator_deregister_no_status);
	tcase_add_test (tc_core, test_validator_deregister_with_status);
	tcase_add_test (tc_core, test_validator_check);
	suite_add_tcase (s, tc_core);

	return s;
}

int main (void)
{
	int number_failed;
	Suite *s;
	SRunner *sr;

	s = test_suite ();
	sr = srunner_create (s);

	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);

	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

