/*
 * test_validator_tag.c
 *
 * Copyright 2023 Jonathan Schöbel <jonathan@xn--schbel-yxa.info>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <check.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>


/* lower SIZE_MAX as we try to reach it */
#include <limits.h>
#undef SIZE_MAX
#define SIZE_MAX 10 * sizeof (struct tag_info)

#include "macro.h"
#include "status.h"

/* override HTML spec */
#include "validator_html.h"
const struct HTML_TAG_DEFINITION HTML_[] = {
	{"html"},
	{"aside"},
	{"html"},
	{"body"}
};
#define HTML5 HTML_

/* C file is needed, because we want to override SIZE_MAX */
#include "validator.c"


START_TEST(test_validator_no_status)
{
	struct SH_Validator * validator;

	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	ck_assert_int_eq (0, validator->tag_n);
	ck_assert_int_eq (TAG_ERR, validator->last_tag);

	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;

	_status_preinit (status);
	validator = SH_Validator_new (&status);
	ck_assert_ptr_ne (NULL, validator);
	ck_assert_int_eq (status.status, SUCCESS);

	ck_assert_int_eq (0, validator->tag_n);
	ck_assert_int_eq (TAG_ERR, validator->last_tag);

	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_copy_no_status)
{
	struct SH_Validator * validator;
	struct SH_Validator * copy;

	/* setup */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test */
	copy = SH_Validator_copy (validator, NULL);
	ck_assert_ptr_ne (NULL, copy);

	ck_assert_ptr_ne (NULL, copy->tags);
	ck_assert_ptr_ne (validator->tags, copy->tags);
	ck_assert_int_eq (validator->tag_n, copy->tag_n);
	ck_assert_int_eq (validator->last_tag, copy->last_tag);

	for (size_t index = 0; index < copy->tag_n; index++)
	{
		ck_assert_int_eq (validator->tags[index].id,
		                  copy->tags[index].id);
		ck_assert_ptr_ne (validator->tags[index].name,
		                  copy->tags[index].name);
		ck_assert_str_eq (validator->tags[index].name,
		                  copy->tags[index].name);
	}

	SH_Validator_free (copy);
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_copy_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;
	struct SH_Validator * copy;

	/* setup */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test */
	_status_preinit (status);
	copy = SH_Validator_copy (validator, &status);
	ck_assert_ptr_ne (NULL, copy);
	ck_assert_int_eq (status.status, SUCCESS);

	ck_assert_ptr_ne (NULL, copy->tags);
	ck_assert_ptr_ne (validator->tags, copy->tags);
	ck_assert_int_eq (validator->tag_n, copy->tag_n);
	ck_assert_int_eq (validator->last_tag, copy->last_tag);

	for (size_t index = 0; index < copy->tag_n; index++)
	{
		ck_assert_int_eq (validator->tags[index].id,
		                  copy->tags[index].id);
		ck_assert_ptr_ne (validator->tags[index].name,
		                  copy->tags[index].name);
		ck_assert_str_eq (validator->tags[index].name,
		                  copy->tags[index].name);
	}

	SH_Validator_free (copy);
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_spec_no_status)
{
	struct SH_Validator * validator;

	/* test */
	validator = SH_Validator_new_html5 (NULL);
	ck_assert_ptr_ne (NULL, validator);

	ck_assert_ptr_ne (NULL, validator->tags);
	ck_assert_int_eq (3, validator->tag_n);
	ck_assert_int_eq (3, validator->last_tag);

	ck_assert_int_eq (1, validator->tags[0].id);
	ck_assert_ptr_ne (HTML_[1].tag, validator->tags[0].name);
	ck_assert_str_eq (HTML_[1].tag, validator->tags[0].name);

	ck_assert_int_eq (2, validator->tags[1].id);
	ck_assert_ptr_ne (HTML_[3].tag, validator->tags[1].name);
	ck_assert_str_eq (HTML_[3].tag, validator->tags[1].name);

	ck_assert_int_eq (3, validator->tags[2].id);
	ck_assert_ptr_ne (HTML_[0].tag, validator->tags[2].name);
	ck_assert_str_eq (HTML_[0].tag, validator->tags[2].name);

	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_spec_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;

	/* test */
	_status_preinit (status);
	validator = SH_Validator_new_html5 (&status);
	ck_assert_ptr_ne (NULL, validator);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_ptr_ne (NULL, validator->tags);
	ck_assert_int_eq (3, validator->tag_n);
	ck_assert_int_eq (3, validator->last_tag);

	ck_assert_int_eq (1, validator->tags[0].id);
	ck_assert_ptr_ne (HTML_[1].tag, validator->tags[0].name);
	ck_assert_str_eq (HTML_[1].tag, validator->tags[0].name);

	ck_assert_int_eq (2, validator->tags[1].id);
	ck_assert_ptr_ne (HTML_[3].tag, validator->tags[1].name);
	ck_assert_str_eq (HTML_[3].tag, validator->tags[1].name);

	ck_assert_int_eq (3, validator->tags[2].id);
	ck_assert_ptr_ne (HTML_[0].tag, validator->tags[2].name);
	ck_assert_str_eq (HTML_[0].tag, validator->tags[2].name);

	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

START_TEST(test_validator_tag_register_no_status)
{
	struct SH_Validator * validator;
	const char * tag1 = "html";
	const char * tag2 = "head";
	const char * tag3 = "article";
	const char * tag4 = "p";
	const char * tag5 = "img";
	char * tagN;
	Tag tag;
	Tag tag_;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test - register */
	tag = SH_Validator_register_tag (validator, tag1, NULL);
	ck_assert_int_eq (1, tag);

	ck_assert_int_eq (validator->tag_n, 1);
	ck_assert_int_eq (validator->last_tag, tag);

	ck_assert_int_eq (validator->tags[0].id, tag);
	ck_assert_str_eq (validator->tags[0].name, tag1);

	/* test - duplicate registration */
	tag_ = SH_Validator_register_tag (validator, tag1, NULL);
	ck_assert_int_eq (tag_, tag);

	ck_assert_int_eq (validator->tag_n, 1);
	ck_assert_int_eq (validator->last_tag, tag);

	ck_assert_int_eq (validator->tags[0].id, tag);
	ck_assert_str_eq (validator->tags[0].name, tag1);

	/* test - order */
	tag = SH_Validator_register_tag (validator, tag3, NULL);
	ck_assert_int_eq (tag, 2);

	tag = SH_Validator_register_tag (validator, tag4, NULL);
	ck_assert_int_eq (tag, 3);

	tag = SH_Validator_register_tag (validator, tag5, NULL);
	ck_assert_int_eq (tag, 4);

	ck_assert_int_eq (validator->tag_n, 4);

	ck_assert_str_eq (validator->tags[0].name, tag3);
	ck_assert_str_eq (validator->tags[1].name, tag1);
	ck_assert_str_eq (validator->tags[2].name, tag5);
	ck_assert_str_eq (validator->tags[3].name, tag4);

	/* test - overflow detection */
	/* make method fail by filling with garbage until
	 * upper boundary is reached */

	/* ensure enough space  inside string*/
	/* log10 +1 = number length */
	/* +3 "tag" */
	/* +1 '\0' */
	/* = +5 */
	tagN = calloc (((int) floor (log10 ((double) SIZE_MAX))) + 5,
		       sizeof (char));

	/* fill with garbage */
	do
	{
		sprintf (tagN, "tag%zu", validator->tag_n);
	}
	while (TAG_ERR != SH_Validator_register_tag (validator, tagN, NULL));

	free (tagN);

	/* test overflow #1 */
	tag = SH_Validator_register_tag (validator, tag2, NULL);
	ck_assert_int_eq (TAG_ERR, tag);

	ck_assert_int_eq (validator->tag_n, 10);

	/* test overflow #2 */
	validator->tag_n = 1;
	validator->last_tag = TAG_MAX;

	tag = SH_Validator_register_tag (validator, tag2, NULL);
	ck_assert_int_eq (TAG_ERR, tag);

	ck_assert_int_eq (validator->tag_n, 1);
	ck_assert_int_eq (validator->last_tag, TAG_MAX);

	/* cleanup */
	/* also free garbage created for overflow test */
	validator->tag_n = 10;
	SH_Validator_free (validator);
}

START_TEST(test_validator_tag_register_with_status)
{
	struct SH_Status status;
	struct SH_Validator * validator;
	const char * tag1 = "html";
	const char * tag2 = "head";
	const char * tag3 = "article";
	const char * tag4 = "p";
	const char * tag5 = "img";
	char * tagN;
	Tag tag;
	Tag tag_;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	/* test - register */
	_status_preinit (status);
	tag = SH_Validator_register_tag (validator, tag1, &status);
	ck_assert_int_eq (1, tag);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (validator->tag_n, 1);
	ck_assert_int_eq (validator->last_tag, tag);

	ck_assert_int_eq (validator->tags[0].id, tag);
	ck_assert_str_eq (validator->tags[0].name, tag1);

	/* test - duplicate registration */
	_status_preinit (status);
	tag_ = SH_Validator_register_tag (validator, tag1, &status);
	ck_assert_int_eq (tag_, tag);
	ck_assert_int_eq (SUCCESS, status.status);

	ck_assert_int_eq (validator->tag_n, 1);
	ck_assert_int_eq (validator->last_tag, tag);

	ck_assert_int_eq (validator->tags[0].id, tag);
	ck_assert_str_eq (validator->tags[0].name, tag1);

	/* test - order */
	_status_preinit (status);
	tag = SH_Validator_register_tag (validator, tag3, &status);
	ck_assert_int_eq (tag, 2);
	ck_assert_int_eq (status.status, SUCCESS);

	_status_preinit (status);
	tag = SH_Validator_register_tag (validator, tag4, &status);
	ck_assert_int_eq (tag, 3);
	ck_assert_int_eq (status.status, SUCCESS);

	_status_preinit (status);
	tag = SH_Validator_register_tag (validator, tag5, &status);
	ck_assert_int_eq (tag, 4);
	ck_assert_int_eq (status.status, SUCCESS);

	ck_assert_int_eq (validator->tag_n, 4);

	ck_assert_str_eq (validator->tags[0].name, tag3);
	ck_assert_str_eq (validator->tags[1].name, tag1);
	ck_assert_str_eq (validator->tags[2].name, tag5);
	ck_assert_str_eq (validator->tags[3].name, tag4);

	/* test - overflow detection */
	/* make method fail by filling with garbage until
	 * upper boundary is reached */

	/* ensure enough space  inside string*/
	/* log10 +1 = number length */
	/* +3 "tag" */
	/* +1 '\0' */
	/* = +5 */
	tagN = calloc (((int) floor (log10 ((double) SIZE_MAX))) + 5,
		       sizeof (char));

	/* fill with garbage */
	do
	{
		sprintf (tagN, "tag%zu", validator->tag_n);
	}
	while (TAG_ERR != SH_Validator_register_tag (validator, tagN, NULL));

	free (tagN);

	/* test overflow #1 */
	_status_preinit (status);
	tag = SH_Validator_register_tag (validator, tag2, &status);
	ck_assert_int_eq (TAG_ERR, tag);
	ck_assert_int_eq (E_DOMAIN, status.status);

	ck_assert_int_eq (validator->tag_n, 10);

	/* test overflow #2 */
	validator->tag_n = 1;
	validator->last_tag = TAG_MAX;

	_status_preinit (status);
	tag = SH_Validator_register_tag (validator, tag2, &status);
	ck_assert_int_eq (TAG_ERR, tag);
	ck_assert_int_eq (E_DOMAIN, status.status);

	ck_assert_int_eq (validator->tag_n, 1);
	ck_assert_int_eq (validator->last_tag, TAG_MAX);

	/* cleanup */
	/* also free garbage created for overflow test */
	validator->tag_n = 10;
	SH_Validator_free (validator);
}

START_TEST(test_validator_tag_check)
{
	struct SH_Validator * validator;
	const char * tag1  = "html";
	const char * tag2  = "html";
	const char * tag3  = "head";
	Tag tag;
	bool result;

	/* setup */
	validator = SH_Validator_new (NULL);
	ck_assert_ptr_ne (NULL, validator);

	tag = SH_Validator_register_tag (validator, tag1, NULL);
	ck_assert_int_ne (TAG_ERR, tag);

	/* test */
	result = SH_Validator_check_tag (validator, tag1);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_check_tag (validator, tag2);
	ck_assert_int_eq (TRUE, result);

	result = SH_Validator_check_tag (validator, tag3);
	ck_assert_int_eq (FALSE, result);

	/* cleanup */
	SH_Validator_free (validator);
}
END_TEST

Suite * test_suite (void)
{
	Suite *s;
	TCase *tc_core;

	s = suite_create ("Testsuite SeFHT Validator Tag");

	/* Core test case */
	tc_core = tcase_create ("Core");

	tcase_add_test (tc_core, test_validator_no_status);
	tcase_add_test (tc_core, test_validator_with_status);
	tcase_add_test (tc_core, test_validator_copy_no_status);
	tcase_add_test (tc_core, test_validator_copy_with_status);
	tcase_add_test (tc_core, test_validator_spec_no_status);
	tcase_add_test (tc_core, test_validator_spec_with_status);
	tcase_add_test (tc_core, test_validator_tag_register_no_status);
	tcase_add_test (tc_core, test_validator_tag_register_with_status);
	tcase_add_test (tc_core, test_validator_tag_check);
	suite_add_tcase (s, tc_core);

	return s;
}

int main (void)
{
	int number_failed;
	Suite *s;
	SRunner *sr;

	s = test_suite ();
	sr = srunner_create (s);

	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);

	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

